#!/usr/bin/env python
from math import floor
from sys import argv
from Cobalt.Util import print_tabular

def build_range(ilist):
    '''Build range of jobids'''
    r_issues = []
    extras = []
    for i in ilist:
        try:
            r_issues.append(int(i))
        except ValueError:
            extras.append(i)
    #print issues
    extras.sort()
    r_issues.sort()
    r_issues.reverse()
    ranges = []
    if r_issues:
        current = r_issues.pop()
        ends = [current, current]
        while r_issues:
            new = r_issues.pop()
            if new == (current + 1):
                ends[1] = new
            else:
                if ends[0] != ends[1]:
                    ranges.append("%s-%s"%tuple(ends))
                else:
                    ranges.append(str(ends[0]))
                ends = [new, new]
            current = new
        if ends[0] != ends[1]:
            ranges.append("%s-%s" % tuple(ends))
        else:
            ranges.append(str(ends[0]))
    return ",".join(ranges + extras)


def timeconv(seconds):
    '''Change seconds in : delimited time'''
    t = int(seconds)
    d = int(floor(t/86400))
    t -= (d * 86400)
    h = int(floor(t / 3600))
    t -= (h * 3600)
    m = int(floor(t / 60))
    return "%d:%02d:%02d" % (d, h, m)

if __name__ == '__main__':
    if len(argv) < 2:
        print "Usage: print-usage.py <file1> <file2> ..."

    jobdata = []
    for jobfile in argv[1:]:
        jobdata += open(jobfile, 'r').readlines()

    results = {}
        
    for line in jobdata:
        jobinfo = line.split(' ')
        if '' in jobinfo:
            jobinfo.remove('')
        jobuser = jobinfo[6]
        usertime = jobinfo[12].split(':')[1][:-2]
        nodes = jobinfo[8]
        jobid, user = jobuser.split('/')
        if results.has_key(user):
            results[user][1] += (float(usertime) * int(nodes))
            results[user][0] += [jobid]
        else:
            results[user] = [[jobid], (float(usertime) * int(nodes))]

    for user in results.keys():
        jobs = [int(jobid) for jobid in results[user][0]]
        jobs.sort()
        joblist = build_range(jobs)
        results[user][0] = joblist

    header = [('User', 'Node Minutes', 'JobIDs')]
    data = [(key, timeconv(value[1]), value[0]) for key, value in results.iteritems()]
    print_tabular(header + data)
