#!/usr/bin/env python

'''Cobalt job administration command'''
__revision__ = '$Revision: 1.1 $'

from elementtree.ElementTree import Element, tostring
from getopt import getopt, GetoptError
from sys import argv
from sss.ssslib import comm_lib

helpmsg = 'Usage: cqadm [-d] [--hold] [--release] [--queue=queuename] <jobid> <jobid>'

if __name__ == '__main__':
    try:
        (opts, args) = getopt(argv[1:], 'd', ['hold', 'release', 'queue='])
    except GetoptError, msg:
        print msg
        print helpmsg
        raise SystemExit, 1

    if len(args) == 0:
        print "At least one jobid must be supplied"
        print helpmsg
        raise SystemExit, 1

    if ('-d', '') in opts:
        debug = True
        opts.remove(('-d', ''))
    else:
        debug = False

    if len(opts) == 0:
        print "At least one command must be specified"
        print helpmsg
        raise SystemExit, 1

    if ((('--hold', '') in opts) and (('--release', '') in opts)):
        print "Only one of --hold or --release can be used at once"
        print helpmsg
        raise SystemExit, 1

    msg = Element('set-job')
    
    if ('--hold', '') in opts:
        msg.attrib['state'] = 'hold'
    elif ('--release', '') in opts:
        msg.attrib['state'] = 'ready'

    if '--queue' in [opt[0] for opt in opts]:
        [queue] = [opt[1] for opt in opts if '--queue' == opt[0]]
        msg.attrib['queue'] = queue

    for jobid in args:
        msg.append(Element("Job", jobid=jobid))

    comm = comm_lib(debug=debug)
    try:
        handle = comm.ClientInit('queue-manager')
    except:
        print "Failed to connect to queue manager"
        raise SystemExit, 1

    comm.SendMessage(handle, tostring(msg))
    response = comm.RecvMessage(handle)
    comm.ClientClose(handle)
    print response
    
