#!/usr/bin/env python

'''Cobalt Queue Status'''
__revision__ = '$Revision: 1.5 $'

from sss.ssslib import comm_lib, LookupError as SSSLookupError
from getopt import getopt, GetoptError
from math import floor
from sys import argv
from elementtree.ElementTree import XML
from Cobalt.Util import print_tabular

if __name__ == '__main__':
    try:
        (opts, args) = getopt(argv[1:], 'df', ['version'])
    except GetoptError, msg:
        print "Usage: cqstat [--version] [-d] [-f jobid]"
        print msg
        raise SystemExit, 1

    if '--version' in argv:
        print "cqstat %s" % __revision__
        raise SystemExit, 0

    jobid = None

    if len(args) > 0:
        jobid = args[0]

    if jobid:
        jobid_tosend = jobid
    else:
        jobid_tosend = "*"

    comm = comm_lib(debug=('-d' in argv))

    try:
        handle = comm.ClientInit("queue-manager")
    except SSSLookupError:
        print "failed to connect to queue manager"
        raise SystemExit, 1
    if '-f' in argv:
        msg = "<get-job><job user='*' walltime='*' nodes='*' state='*' jobid='%s' location='*' mode='*' procs='*' queue='*'/></get-job>"
    else:
        msg = "<get-job><job user='*' walltime='*' nodes='*' state='*' jobid='%s' location='*'/></get-job>" 
    comm.SendMessage(handle, msg % jobid_tosend)
    xdata = comm.RecvMessage(handle)
    comm.ClientClose(handle)
    resp = XML(xdata)

    if '-f' in argv:
        header = [('JobID', 'User', 'WallTime', 'Nodes', 'State', 'Location', 'Mode', 'Procs', 'Queue')]
    else:
        header = [('JobID', 'User', 'WallTime', 'Nodes', 'State', 'Location')]
        
    output = [[job.get(x) for x in [y.lower() for y in header[0]]] for job in resp.findall('.//job')]

    # next we cook walltime
    for i in xrange(len(output)):
        t = int(output[i][2].split('.')[0])
        h = int(floor(t/60))
        t -= (h * 60)
        output[i][2] = "%02d:%02d:00" % (h, t)

    output.sort()
    print_tabular([tuple(x) for x in header + output])
