#!/usr/bin/env python

'''Cobalt qsub command'''
__revision__ = '$Revision: 1.12 $'

from sys import argv
from os import environ, getcwd, getuid
from pwd import getpwuid
from elementtree.ElementTree import Element, SubElement, XML, tostring
from sss.ssslib import comm_lib
from Cobalt.Util import dgetopt

helpmsg = "Usage: cqsub [-d] [-v] -p <project> -q <queue> -C <working directory> -e envvar1=value1:envvar2=value2 -k <kernel profile> -O <outputprefix> -t time <in minutes> -n <number of nodes> -c <processor count> -m <mode co/vn> <command> <args>"

if __name__ == '__main__':
    if '--version' in argv:
        print "cqsub %s" % __revision__
        raise SystemExit, 0
    options = {'v':'verbose', 'd':'debug'}
    doptions = {'n':'nodecount', 't':'time', 'p':'project', 'm':'mode', 'c':'proccount', 'C':'cwd',
                'e':'env', 'k':'kernel', 'q':'queue', 'O':'outputprefix'}
    (opts, command) = dgetopt(argv[1:], options, doptions, helpmsg)
    # need to filter here for all args
    failed = False
    needed = ['time', 'nodecount'] #, 'project']
    if [field for (field, value) in opts.iteritems() if not value and field in needed] or not command:
        for ofield in needed:
            if opts[ofield]:
                needed.remove(ofield)
        if command:
            print "Not all required arguments provided: %s needed" % (",".join(needed))
        else:
            print "Command required"
        print helpmsg
        raise SystemExit, 1
    try:
        nc = int(opts['nodecount'])
    except:
        print "non-integer node count specified"
        raise SystemExit, 1
    
    if not 0 < nc <= 1024:
        print "node count out of realistic range"
        raise SystemExit, 1
    if opts['cwd'] == False:
        opts['cwd'] = getcwd()
    # ensure time is actually in minutes
    if opts['time'].count(':') > 0:
        # process as a time
        #print "assuming seconds are not included in %s" % opts['time']
        units = opts['time'].split(':')
        units.reverse()
        # kill seconds
        units.pop()
        totaltime = 0
        mults = [60, 1]
        if len(units) > 3:
            print "time too large"
            raise SystemExit, 1
        totaltime = sum([mults[index] * float(units[index]) for index in range(len(units))])
        print "submitting walltime=%s minutes" % str(totaltime)
        opts['time'] = str(totaltime)
    if float(opts['time']) <= 0:
        print "invalid time specification"
        raise SystemExit, 1
    user = getpwuid(getuid())[0]
    if command[0][0] != '/':
        command[0] = opts['cwd'] + '/' + command[0]
    request = Element('add-job')
    if not opts['mode']:
        opts['mode'] = 'co'
    for field in ['kernel', 'queue']:
        if not opts[field]:
            opts[field] = 'default'
    if not opts['proccount']:
        if opts.get('mode', 'co') == 'vn':
            # set procs to 2 x nodes
            opts['proccount'] = str(2 * int(opts['nodecount']))
        else:
            opts['proccount'] = opts['nodecount']
    job = SubElement(request, "job", nodes=opts['nodecount'], #project=opts['project'],
                     user=user, outputdir=opts['cwd'], walltime=opts['time'], path=environ['PATH'],
                     mode=opts.get('mode', 'co'), procs=opts.get('proccount'), kernel=opts['kernel'],
                     queue=opts['queue'])
    if opts['outputprefix']:
        if opts['outputprefix'][0] == '/':
            job.set('outputpath', "%s.output" % (opts['outputprefix']))
            job.set('errorpath', "%s.error" % (opts['outputprefix']))
        else:
            job.set('outputpath', "%s/%s.output" % (opts['cwd'], opts['outputprefix']))
            job.set('errorpath', "%s/%s.error" % (opts['cwd'], opts['outputprefix']))
    if opts['env']:
        [SubElement(job, "env", name=data.split('=')[0], value=data.split('=')[1]) for data in opts['env'].split(':')]
    #for (key, value) in environ.iteritems():
    #    SubElement(job, "env", name=key, value=value)
    cmd = SubElement(job, "command", bin=command[0], args=" ".join(command[1:]))
    #print tostring(request)
    
    comm = comm_lib(debug=opts['debug'])
    handle = comm.ClientInit("queue-manager")
    comm.SendMessage(handle, tostring(request))
    response = comm.RecvMessage(handle)
    comm.ClientClose(handle)

    try:
        jobid = XML(response).find('job').attrib['jobid']
        print jobid
    except:
        print "Error submitting job"
        raise SystemExit, 1
