#!/usr/bin/env python

from sss.ssslib import comm_lib
from getopt import getopt, GetoptError
from sys import argv, exit, stdout, stdin
from string import join,strip
from elementtree.ElementTree import Element, SubElement, XML, tostring
from xml.sax.saxutils import escape
from select import select
from signal import signal, SIGINT
from socket import socket,AF_INET,SOCK_STREAM,gethostname

import pwd
import os

def ignore(signum,frame):
    pass

def parse_scriptfile(filename):
    walltime = None
    nodes = None
    procs = None
    image = None
    stagein = None
    stageout = None
    project = None
    attribute = None
    script = 0
    script_data = ""
    lines = map(strip,open(filename).readlines())

    if lines[0] != "%setup":
        print "First line in script must be %setup"
        exit(1)
    print lines[1:]

    theEnd = 0
    i = 0
    for line in lines[0:]:
        if (theEnd == 0):
            if line.find("%run") > -1:
                theEnd = i
        i=i + 1

    if (theEnd == 0):
        print "You must have a %run section in your script"
        exit(1)
    for line in lines[1:theEnd]:
       # print 'parsing :%s:'%line
        if line.find("nodes") > -1:
            nodes = strip(line.split(":")[1])
        #    print "a"
        elif line.find("procs") > -1:
            procs = strip(line.split(":")[1])
        elif line.find("walltime") > -1:
            walltime = strip(line.split(":")[1])
         #   print "b"
        elif line.find("stagein") > -1:
            stagein = strip(line.split(":")[1])
          #  print "c"
        elif line.find("stageout") > -1:
            stageout = strip(line.split(":")[1])
           # print "d"
        elif line.find("project") > -1:
            project = strip(line.split(":")[1])
            #print "e"
        elif line.find("attribute") > -1:
            attribute = strip(line.split(":")[1])
        elif line.find("image") > -1:
            image = strip(line.split(":")[1])
            #print "f"
            
    for line in lines[(1+theEnd):]:    
        script_data += "%s\n"%line

    if (not walltime) or (not project) or (not nodes):
        print "Must give walltime, project, nodes in script file"
        exit(1)

    return (walltime,nodes,procs,image,stagein,stageout,project,attribute,script_data)

def parsePBSscript(filename):
    shell=None
    lines=map(strip,open(filename).readlines())
    for line in lines:
        if line.find("#PBS -l nodes=") > -1:
            nodes=split(split(line,"=")[1],':')[0]
        elif line.find("#PBS -l walltime=") > -1:
            pbs_walltime = split(line,"=")[1]
            times=map(int,split(pbs_walltime,":"))
            walltime=0
            for i in xrange(len(times)):
                walltime=walltime+(times[-(i+1)]*(60**(i-1)))
            walltime = str(walltime)
        elif line.find("#PBS -A") > -1:
            project = split(line)[2]
        elif line.find("#ctrans in ") > -1:
            stagein = split(line)[2]
        elif line.find("#ctrans out ") > -1:
            stageout = split(line)[2]
        elif line.find("#PBS -S") > 1:
            shell = split(line)[2]
    script_data=join(lines,"\n")
    if script_data[:2] != "#!":
        if not shell:
            shell=getpwent(getuid())[-1]
        script_data="#!%s"%(shell)+script_data

    if (not walltime) or (not nodes) or (not project):
        print "Incomplete pbs script file!  file is missing the following fields:"
        if (not walltime):
            print "walltime"
        if (not nodes):
            print "nodes"
        if (not project):
            print "project"
        exit(1)

    return (walltime,nodes,image,stagein,stageout,project,None,script_data)

if __name__ == '__main__':
    t='mpish'
    if len(argv) > 1 and argv[1] == '-f':
        (walltime,nodes,procs,image,stagein,stageout,project,attribute,script) = parse_scriptfile(argv[2])
    elif '-P' in argv:
        (walltime,nodes,procs,image,stagein,stageout,project,attribute,script) = parsePBSscript(argv[2])
        t='pbs'
    else:
        try:
            (opts,args) = getopt(argv[1:],'PIdc:p:t:n:i:o:m:')
        except GetoptError, msg:
            print msg
            print "Usage: qsub.py [-P] [-d] [-I] -p project -t time (in minutes) -n (number of nodes) -c ( number of Processes ) -a (node attribute) -i (stage in directory) -o (stage out directory) [-m image] <scriptfile>"
            print "Usage:        or"
            print "Usage: qsub.py -f <qsub scriptfile>"
            exit(1)

        walltime = None
        nodes = None
        procs = None
        image = None
        stagein = None
        stageout = None
        project = None
        attribute = None
        lienID = None
        PBS = None

        for opt in opts:
            if opt[0] == '-t':
                walltime = opt[1]
            elif opt[0] == '-n':
                nodes = opt[1]
            elif opt[0] == '-c':
                procs = opt[1]
            elif opt[0] == '-i':
                stagein = opt[1]
            elif opt[0] == '-o':
                stageout = opt[1]
            elif opt[0] == '-m':
                image = opt[1]
            elif opt[0] == '-p':
                project = opt[1]
            elif opt[0] == '-a':
                attribute = opt[1]
            elif opt[0] == '-P':
                PBS=1

        
        if (not walltime) or (not project) or (not nodes) or (not len(args)):
            print "Usage: qsub.py [-d] [-I] -p project -t time (in minutes) -n (number of nodes) -c ( number of processes ) -i (stage in directory) -o (stage out directory) [-m image] <scriptfile>"
            exit(1)

        scriptfile = args[0]
        script = open(scriptfile,"r").read()

    if not procs:
        procs = nodes
        
    user = pwd.getpwuid(os.getuid())[0]

    # Verify that walltime is sane
    wallminutes = None
    try:
        wallminutes = int(walltime)
    except ValueError, e:
        # Check if time was given in colon format
        times = walltime.split(":")
        if len(times) == 2:
            wallminutes = int(times[0]) * 60 + int(times[1])
        elif len(times) == 3:
            wallminutes = int(times[0]) * 1440 + int(times[1]) * 60 + int(times[2])
        else:
            print "Time parameter misunderstood, please give in terms of whole minutes."
            exit(1)

    tosend = Element("add-job")
    job_record = SubElement(tosend, "job", user=str(user), walltime=str(wallminutes), nodes=nodes, procs=procs, type=t, outputdir=os.getcwd(), project=project, homedir=os.getenv("HOME"))

    if project not in os.popen('showp').read().strip().split():
        print "bad project: %s"%(project)
        exit(2)
    
    if stagein:
        job_record.set("stagein",stagein)
        try:
            s = os.stat(stagein)
        except OSError, e:
            print "Bad stagein directory %s\n%s"%(stagein, e[1])
            exit(1)
            
    if stagein and stageout:
       job_record.set("stageout",stageout)
    if attribute:
        job_record.set("attribute",attribute)

    if '-I' in argv:
        sock = socket(AF_INET,SOCK_STREAM)
        sock.bind(('',0))
        sock.listen(5)
        actualPort = sock.getsockname()[1]
        job_record.set("host",gethostname())
        job_record.set("port",str(actualPort))

    if image:
        job_record.set("image",image)

    for (k,v) in os.environ.iteritems():
        SubElement(job_record, "env", name=k, value=v)

    script_xml = SubElement(job_record, "script")
    script_xml.text = escape(script)

    greq = '''<get-project><project name='%s'><user name='%s'/></project></get-project>''' % (project, user)
    areq = '''<add-lien amount='%s'><project name='%s'/></add-lien>''' % (str(float(wallminutes) * 60), project)

    comm = comm_lib(debug=('-d' in argv))
    try:
        am_handle = comm.ClientInit("allocation-manager")
        comm.SendMessage(am_handle, greq)
        r = comm.RecvMessage(am_handle)
    except:
        print "Failed to contact allocation manager"
        raise SystemExit, 1

    if not len(XML(r).getchildren()):
        print "Allocation error for user %s submitting with project %s (project doesn't exist)" % (user, project)
        raise SystemExit, 1

    try:
        comm.SendMessage(am_handle, areq)
        r = comm.RecvMessage(am_handle)
    except:
        print "Failed to contact allocation manager"
        raise SystemExit, 1

    try:
        lienID = XML(r).find('lien').get('id')
        job_record.set("lienID",lienID)
    except:
        print "Allocation error for user %s submitting with project %s (insufficient balance)" % (user, project)
        raise SystemExit, 1

    comm.ClientClose(am_handle)

    qm_handle = comm.ClientInit("queue-manager")
    comm.SendMessage(qm_handle,tostring(tosend))

    resp = comm.RecvMessage(qm_handle)
    comm.ClientClose(qm_handle)

    try:
        jobid = XML(resp).find('job').attrib['jobid']
    except:
        print "Error submitting job"
        exit(1)


    if jobid == "invalid project":
        print "Project %s is invalid for your username."%project
    else:
        print jobid
    if '-I' in argv:
        print "Waiting for connection from interactive job...",
        stdout.flush()

        # disable SIGINT
        signal(SIGINT, ignore)
        conn,info = sock.accept()
        print "Connected!"
        should_exit = 0
        try:
            while not should_exit:
                [reads,foo,bar] = select([stdin,conn],[],[])
                for r in reads:
                    if r == stdin:
                        str = r.readline()
                        if len(str):
                            conn.send(str)
                        else:
                            conn.close()
                            should_exit = 1
                    elif r == conn:
                        str = conn.recv(1)
                        if len(str):
                            stdout.write(str)
                            stdout.flush()
                        else:
                            conn.close()
                            should_exit = 1
        except KeyboardInterrupt:
            conn.close()
