#!/usr/bin/env python

'''Setup reservations in the scheduler'''
__revision__ = '$Id: setres.py 1.4 05/06/14 15:38:51-05:00 desai@topaz.mcs.anl.gov $'

from sss.ssslib import comm_lib
from sys import argv
from getopt import getopt, GetoptError
from elementtree.ElementTree import Element, SubElement, tostring
from time import mktime, strftime, localtime

help = '''Usage: setres -n name -s <starttime> -d <duration> -p <partition> -u <user>
starttime is in format: YYYY_MM_DD-HH:MM
duration may be in minutes or DD:HH:MM
user and name are optional'''

if __name__ == '__main__':
    try:
        (opts, args) = getopt(argv[1:], 's:d:n:p:u:', [])
    except GetoptError, msg:
        print msg
        print help
        raise SystemExit, 1
    try:
        [partition] = [opt[1] for opt in opts if opt[0] == '-p']
        [start] = [opt[1] for opt in opts if opt[0] == '-s']
        [duration] = [opt[1] for opt in opts if opt[0] == '-d']
    except:
        print "Must supply -s, -d, and -p options with values"
        print help
        raise SystemExit, 1
    if duration.count(':') == 0:
        dsec = int(duration) * 60
    else:
        dinfo = duration.split(':')
        dinfo.reverse()
        mults = [60, 3600, 86400]
        dsec = 0
        dsec = sum([int(dinfo[index]) * mults[index] for index in range(len(dinfo))])
    (day, time) = start.split('-')
    (syear, smonth, sday) = [int(field) for field in day.split('_')]
    (shour, smin) = [int(field) for field in time.split(':')]
    starttime = mktime((syear, smonth, sday, shour, smin, 0, 0, 0, -1))
    print "Got starttime %s" % (strftime('%c', localtime(starttime)))
    msg = Element('AddReservation', start=str(starttime), duration=str(dsec))
    if '-u' in argv[1:]:
        msg.set('user', [opt[1] for opt in opts if opt[0] == '-u'][0])
    else:
        msg.set('user', '')
    if '-n' in argv[1:]:
        [nameinfo] = [val for (opt,val) in opts if opt == '-n']
        msg.set('name', nameinfo)
    SubElement(msg, 'Partition', name=partition)
    print tostring(msg)
    comm = comm_lib()
    handle = comm.ClientInit('scheduler')
    comm.SendMessage(handle, tostring(msg))
    resp = comm.RecvMessage(handle)
    comm.ClientClose(handle)
    print resp
