#!/usr/bin/env python
'''this is the anl version of the SSS scheduler'''
__revision__ = "0.0.5"

from sss.restriction import Data, DataSet, ID
from sss.server import EventReceiver 
from sss.daemonize import daemonize
from sss.ssslib import comm_lib, ConnectError, CommunicationError, LookupError as SSSLookupError

from syslog import syslog, LOG_INFO, LOG_ERR
from re import match, compile as regcompile 
import time
from signal import signal, SIGCHLD, SIG_DFL
from elementtree.ElementTree import Element, XML

MAX_DATE = 2147483647

class TimerException(Exception):
    '''a set of exceptions currently unused'''
    pass

class Timer(object):
    '''a generic timer'''
    def __init__(self):
        '''initialize the timer'''
        self.start = time.time()
        self.stop = time.time()
        pass
    
    def Start(self):
        '''start the timer'''
        self.start = time.time()
        
    def Stop(self):
        '''stop the timer'''
        self.stop = time.time()
            
    def Check(self):
        '''check the timer to see the current status'''
        try:
            return self.stop-self.start
        except:
            raise TimerException, "StateError"
        
    def Elapsed(self):
        '''get the elapsed time of the timer'''
        try:
            return time.time()-self.start
        except:
            raise TimerException, "StateError" 
        
class Message:
    '''the object used to store data from the incomming messages'''
    def __init__(self, message, data):
        '''initialize the Message object'''
        self.msg = message
        self.data = data

    def getMessage(self):
        '''return the message'''
        return self.msg

    def getData(self):
        '''return the data'''
        return self.data

class ComputeNode:
    '''object for nodes in the system'''
    def __init__(self, name, attributes):
        '''initialize the object with default values'''
        self.name = name
        self.attributes = attributes.split(',')
        self.status = "offline"
        self.currentjob = ""

    def hasAttribute(self, attribute):
        '''get the nodes attributes'''
        return self.attributes.__contains__(attribute)

    def isAvailable(self):
        '''is the node currently available to be scheduled'''
        return self.status == "online" and self.currentjob == ""

    def isOnline(self):
        '''is the node up and ready to be used'''
        return self.status == "online"
    
class Job(Data):
    '''Object to maintain data about a job'''

    def __init__(self, xml, resid):
        '''Initialize the object with default values'''
        Data.__init__(self, xml)
        self.attr['status'] = "waiting"
        self.queuestatus = ""
        self.type = "reservation"
        if not self.attr.has_key('starttime'):
            self.attr['starttime'] = ""
	else:
	    self.attr['status'] = "scheduled"
        if not self.attr.has_key('attribute'):
            self.attr['attribute'] = "compute"
        if not self.attr.has_key('jobid'):
            self.attr['jobid'] = "SYSTEM.%s"% resid
        self.setNodelist([])

    def getUsers(self):
        '''this will provide a list of users authorized to manipulate jobs'''
        return self.attr['user']

    def getNodelist(self):
        '''provides the list of nodes being used by the job'''
	nodelist =[]
	for nodeelement in self.element.findall("Node"):
		nodelist.append( nodeelement.attrib['name'])
        return nodelist
    
    def setNodelist(self, setlist):
        '''create the list of nodes the job is allotted'''
        for nodeElement in self.element.findall("Node"):
            self.element.remove(nodeElement)
        for node in setlist:
            self.element.append(Element("Node", name = "%s"%node))
        
    def getStartTime(self):
        '''returns the start time of the job'''
        if self.attr['starttime']:
            return float(self.attr['starttime'])
        else:
            return 0
        
    def getEndTime(self):
        '''returns the endtime of the job'''
        if self.attr['walltime']:
            return self.getStartTime() + (60 * float(self.attr['walltime']))
        else:
            return 0
 
    def runsDuring(self, other):
        '''checks the job against another job to see if they will run at the same time'''
        if self.getStartTime() >= other.getStartTime() and self.getStartTime() <= other.getEndTime():
            return 1
        elif self.getEndTime() >= other.getStartTime() and self.getEndTime() <= other.getEndTime():
            return 1
        elif self.getStartTime() < other.getStartTime() and  self.getEndTime() > other.getEndTime():
            return 1
        else:
            return 0

    def Run(self, comm):
        '''actually tell the queue manager to run the job or start the reservation'''
        commastring = ":"
        if self.type == "reservation":
            self.attr['status'] = "running"
	    syslog( LOG_INFO, "just ran reservation: %s"%self.attr['jobid'] )
        else:
            try:
                h = comm.ClientInit("queue-manager")
		syslog( LOG_INFO, "connected to queue-manager to run jobid: %s"%self.attr['jobid'] )
                comm.SendMessage(h, '''<run-job nodelist="%s" ><job jobid="%s"/></run-job>'''% (commastring.join(self.getNodelist()), self.attr['jobid']) )
                runmsg = comm.RecvMessage(h)
                if runmsg:
                    syslog( LOG_INFO, "just sheduled jobid: %s"%self.attr['jobid'] )
                    self.attr['status'] = 'running'
                    comm.ClientClose(h)
            except (ConnectError, CommunicationError, SSSLookupError):
                syslog( LOG_ERR, "couldn't contact the queue-manager to run job %s"% self.attr['jobid'] )
                self.attr['status'] = 'waiting'
                self.setNodelist([])
		syslog( LOG_INFO,"cleared nodelist: here is what it is now: %s"%commastring.join(self.getNodelist()))

        
class Nodelist:
    '''Object to hold the list of nodes that the schedluer will use'''
    def __init__(self, comm, nodefile ):
        '''Initialize the object and create the actual list data'''
        signal(SIGCHLD, SIG_DFL)
        self.list = []
        for line in [ line.strip() for line in open(nodefile, 'r').readlines() ]:
            name, attributes = line.split(":")
            self.list.append(ComputeNode(name, attributes))
        try:
            h = comm.ClientInit("node-state-manager")
            comm.SendMessage( h, "<get-node-state><node-state node='*' adminstate='*' /></get-node-state>")
            nodemsg =  comm.RecvMessage(h)
            comm.ClientClose(h)
            domobj = XML(nodemsg)
            regex = regcompile('(?P<host>ccn\d+)(\.\w+)*')
            for nsmNode in domobj.findall("node-state"):
                mymatch = match(regex, nsmNode.attrib['node'])
                if mymatch:
                    mynode = mymatch.group('host')
                    if nsmNode.attrib['adminstate'] == 'online':
                        self.Activate(mynode)
                    else:
                        self.Deactivate(mynode)
        except (ConnectError, CommunicationError, SSSLookupError):
            syslog( LOG_ERR, "couldn't contact the node-state-manager nodelist may be out of sync" )
        #this is the mpd sync that I should probably move or dump in favor of the NSM
        #regex = regcompile('(?P<host>ccn\d+)')
        #regex = regcompile('(?P<host>\w+)')
        #for node in [ mpdnode.strip() for mpdnode in popen(program).readlines() ]:
        #    mymatch = match( regex, node )
        #    if mymatch:
        #        self.Activate(node)

    def matchNodes(self, regex):
        '''create a node list for the job from a regular expression'''
        nodelist = []
        noderegex = regcompile(regex)
        for node in self.list:
            mymatch = match(noderegex, node.name)
            if mymatch:
                nodelist.append(node.name)
        return nodelist

    def maxNodes(self):
        '''return the maximum number of nodes available to the scheduler'''
        numOfNodes = 0
        for node in self.list:
            if node.isOnline():
                numOfNodes = numOfNodes +1
        return numOfNodes
    
    def enoughNodes(self, numOfNodes, attribute, nodelist ):
        '''returns a list of nodes if there are enough nodes available to be scheduled'''
        mynodelist = []
        listofnodes = []
        for node in self.list:
            if node.isAvailable() and node.hasAttribute(attribute) and node.name not in nodelist:
	    	#syslog( LOG_INFO, "node: %s is available to be used by a job"%node.name)
                mynodelist.append(node.name)
        if len(mynodelist) >= numOfNodes:
            listofnodes = mynodelist[:numOfNodes]
	else:
	    syslog( LOG_INFO, "nodes requested: %s  Attribute requested: %s  Nodes available with attribute: %s"%(numOfNodes,attribute,len(mynodelist)))
        return listofnodes

    def areStillAvailable(self, nodelist):
        '''make sure that nodes are still available to be used when it is time for the job to run'''
        counter = 0
        for node in nodelist:
            for mynode in self.list:
                if node == mynode.name:
		    syslog(LOG_INFO,"checking node: %s for availability"%node)
                    if mynode.isAvailable():
		        syslog( LOG_INFO, "node: %s is available to be used by a job"%node)
                        counter = counter +1
                    else:
                        syslog( LOG_INFO, "try to run job but node: %s is not available"%mynode.name)
                        return 0
        if counter == len(nodelist):
	    syslog(LOG_INFO,"ALL NODES AVAILABLE")
            return 1
        return 0
     
    
    def Activate(self, name):
        '''activate a node in the nodelist'''
        self.updateStatus(name, "online")

    def Deactivate(self, name):
        '''deactivate a node in the nodelist'''
        self.updateStatus(name, "offline")

    def updateStatus(self, name, value):
        '''change the status of the node in the node list'''
        for node in self.list:
            if name == node.name:
                node.status = value
                if value == "offline" and node.currentjob:
                    syslog( LOG_ERR, "Node %s is down but still part of Job %s"%(node.name, node.currentjob))

    def markUsed(self, job, nodelist):
        '''mark the node as used by a job or reservation'''
	#syslog( LOG_INFO, "markUsed called by job: %s"%job )
        for node in self.list:
	    for mynode in nodelist:
           	if node.name == mynode:
	    		#syslog( LOG_INFO, "node: %s to be used by a job: %s"%(mynode,job))
                	node.currentjob = job
		

    def markFree(self, nodelist):
        '''mark the node as free or available to be scheduled'''
        for node in self.list:
            if node.name in nodelist:
                node.currentjob = ""
                                            
class Scheduler(EventReceiver):
    '''Core Object for the actual scheduler'''
    __implementation__ = 'anl-sched'
    __component__  =  'scheduler'
    __dispatch__ = {'events': "HandleXMLEvent",
                    'get-schedule':'GetSchedule', #new unified client backend
                    'get-reservations':'reservations.Get',
                    'del-reservations':'reservations.Del',
                    'add-reservations':'reservations.Add'}
    __validate__ = 0
    __subscriptions__ = [('queue-manager', '**', '**'), ('node-state-manager', 'state-changed', '**')]
    __statefields__ = ['reservations']

    def __init__(self,debug):
        '''setup the object with default values'''
	self.comm = comm_lib(debug=debug)
        self.getJobMsg = "<get-job><job jobid='*' state='*' attribute='*' user='*' project='*' nodes='*' walltime='*'/></get-job>"
        self.getNewJobMsg =  "<get-job><job jobid='%s' attribute='*' user='*' project='*' nodes='*' walltime='*'/></get-job>"
        self.getNodeStateMsg =  "<get-node-state><node-state node='%s' state='*' adminstate='*' /></get-node-state>"
        self.adminlist = ['bradshaw', 'desai', 'alusk', 'root']
        self.fifo = []
        self.reservations = DataSet("jobs", "job", Job, ID(), False)
        self.syncTimer = Timer()
        self.tempres = []
        self.deleteme = []
        self.mynodes = Nodelist(self.comm, "/etc/scheduler-nodes")
        self.pendingMsg = []
	EventReceiver.__init__(self)

        
    def __setup__(self):
        self.QueueSync()
        
    def __progress__(self):
        '''method that is the core loop of the daemon'''
        if len(self.pendingMsg):
            self.ProcessMsg()
        if self.syncTimer.Elapsed() > 600:
            self.QueueSync()
            self.syncTimer.Start()
        self.Schedule()
        self.BuildFIFO()
        return 1

    def HandleXMLEvent(self, xml, (peer, port)):
        '''collects all the incoming messages to be processed out of band'''
        for event in xml.findall("event"):
            self.pendingMsg.append(Message(event.attrib['msg'], event.attrib['data']))            
        return Element('event-ok')

    def QueueSync(self):
        '''this provides the ability to make sure the scheduler knows about all the jobs that the queue manager does
           mainly this is just a bit of fault tolerence to insure things don't get lost in the system during restarts'''
        try:
            h = self.comm.ClientInit("queue-manager")
            self.comm.SendMessage( h, self.getJobMsg )
            queuemsg =  self.comm.RecvMessage(h)
            self.comm.ClientClose(h)
            domobj = XML(queuemsg)
            self.tempres = []
            for obj in domobj.findall('job'):
                newjob = Job(obj, id)
                newjob.type = "job"
                newjob.attr['jobid'] = obj.attrib['jobid']
                self.tempres.append(newjob)
            #run comparison
            for rjob in self.reservations:
                if [ res for res in self.tempres if res.attr['jobid'] == rjob.attr['jobid'] ]:
                    for matching in [ res for res in self.tempres if res.attr['jobid'] == rjob.attr['jobid'] ]:
                        self.tempres.remove(matching)
                elif rjob.type == "job":
                    self.mynodes.markFree(rjob.getNodelist())
                    self.reservations.remove(rjob)
                    syslog( LOG_INFO, "I am removing job: %s because the queue-manager doesn't know about it"%
                            rjob.attr['jobid'])
            for qjob in self.tempres:
                if qjob.attr['state'] in [ "ready", "queued", "stage-pending" ]:
                    self.reservations.append(qjob)
                else:
                    syslog( LOG_ERR, "job %s running but I do not have any info about it"
                            %qjob.attr['jobid'] )
        except (ConnectError, CommunicationError, SSSLookupError):
            syslog(LOG_ERR, "Failure to contact the queue-manager")
            
        
                            
    def GetSchedule(self, xe, (peer, port)):
        '''reports the current schedule back to the users showq command'''
        schedMsg = self.reservations.Get(xe, (peer, port))
        schedMsg.attrib['totalnodes'] = str(self.mynodes.maxNodes())
        return schedMsg

    def ProcessMsg(self):
        '''the out of band part of handleXMLmeassage where the guts of message processing happens'''
        while self.pendingMsg:
            event = self.pendingMsg.pop()
            if event.msg == "job-submitted":
                jobid = event.data
                try:
                    h = self.comm.ClientInit("queue-manager")
                    self.comm.SendMessage( h, self.getNewJobMsg% jobid )
                    qmMsg =  self.comm.RecvMessage(h)
                    self.comm.ClientClose(h)
                    domobj = XML(qmMsg)
                    for obj in domobj.findall("job"):
                        newjob = Job(obj, id)
                        newjob.type = "job"
                        self.reservations.append(newjob)
                except (ConnectError, CommunicationError, SSSLookupError):
                    syslog( LOG_ERR, "couldn't contact the queue-manager" )
                    self.pendingMsg.append(event)
            if event.msg == "job-completed" or event.msg == "job-deleted":
                for job in [ curRes for curRes in self.reservations if event.data == curRes.attr['jobid'] ]:
                    syslog( LOG_INFO, "Completed going to remove job: %s"%job.attr['jobid'] )
                    self.mynodes.markFree(job.getNodelist())
                    self.reservations.remove(job)
		for job in [ curRes for curRes in self.reservations if curRes.attr['status'] == "scheduled" ]:
		    job.attr['status'] = "waiting"
            if event.msg == "job-state-changed":
                for affectedjob in [ changed for changed in self.reservations if event.data == changed.attr['jobid'] ]:
                    try:
                        h = self.comm.ClientInit("queue-manager")
                        self.comm.SendMessage( h, "<get-job><job jobid='%s' state='*'/></get-job>"%event.data )
                        jobChangeMsg =  self.comm.RecvMessage(h)
                        self.comm.ClientClose(h)
                        domobj = XML(jobChangeMsg)
                        for obj in domobj.findall("job"):
                            affectedjob.queuestatus = obj.attrib['state']
                    except (ConnectError, CommunicationError, SSSLookupError):
                        syslog( LOG_ERR, "couldn't contact the queue-manager" )
                        self.pendingMsg.append(event)
            if event.msg == "state-changed":
                regex = regcompile('(?P<host>ccn\d+)(\.\w+)*')
                mymatch = match(regex, event.data)
                if mymatch:
                    mynode = mymatch.group('host')
                    node = event.data
                    try:
                        h = self.comm.ClientInit("node-state-manager")
                        self.comm.SendMessage(h, self.getNodeStateMsg%node )
                        nsmMsg = self.comm.RecvMessage(h)
                        self.comm.ClientClose(h)
                        domobj = XML(nsmMsg)
                        for changednode in domobj.findall("node-state"):
                            if changednode.attrib['adminstate'] == "online":
                                syslog( LOG_INFO, "activated %s"%mynode)
                                self.mynodes.Activate(mynode)
                            else:
                                syslog( LOG_INFO, "deactivated %s"%mynode)
                                self.mynodes.Deactivate(mynode)
                    except (ConnectError, CommunicationError, SSSLookupError):
                        syslog( LOG_ERR, "couldn't contact the node-state-manager" )
                        self.pendingMsg.append(event)
            
    def Schedule(self):
        '''act on the current fifo list and do what needs to be done'''
        NOW = time.time()
        for reservation in self.fifo:   # only run through the fifo not the reservation list
            if reservation.getStartTime():
                if reservation.getEndTime() < NOW and reservation.type == "reservation":
                    syslog( LOG_INFO , "reservation: %s is over"%reservation.attr['jobid'])
                    joblist = [ res for res in self.reservations if reservation.attr['jobid'] == res.attr['jobid'] ] 
                    for job in joblist:
			#I need to free nodes
                        self.reservations.remove(job)
                if reservation.getStartTime() <= NOW and reservation.attr['status'] == "scheduled":
		    syslog( LOG_INFO, "job: %s should run"%reservation.attr['jobid'] )
                    if self.mynodes.areStillAvailable(reservation.getNodelist()) or reservation.type == "reservation":
		        syslog( LOG_INFO, "job: %s calling run function"%reservation.attr['jobid'] )
                        reservation.Run(self.comm)
                        self.mynodes.markUsed(reservation.attr['jobid'], reservation.getNodelist())
                    else:
		    	syslog(LOG_INFO,"nodes not avaialble for jobid: %s"%reservation.attr['jobid'] )
		    	reservation.attr['status'] = "waiting"
                        reservation.setNodelist([])
                        
    def BuildFIFO(self):
        '''build the FIFO or actual schedule for the current jobs and reservations in the scheduler'''
	commastring = ':'
        self.fifo = []
        for reservation in self.reservations:
            if reservation.type == "reservation" or reservation.attr['status'] == "running" or reservation.attr['status'] == "scheduled" :
                #if not len(reservation.getNodelist()): # if it doesn't already have a nodelist make one for it.
	        if reservation.attr.has_key('noderegex'):
		        #syslog( LOG_INFO, "job: %s is getting a nodelist"%reservation.attr['jobid'] )
                        reservation.setNodelist(self.mynodes.matchNodes(reservation.attr['noderegex'])) 
                        reservation.attr['nodes'] = str(len(reservation.getNodelist()))
		if reservation.attr['status'] == "running": 
			self.mynodes.markUsed(reservation.attr['jobid'], reservation.getNodelist())
                self.fifo.append(reservation)
        for reservation in self.reservations:    #second pass to insert the jobs
            if reservation.attr['status'] == "waiting" and reservation.type == "job" : #job not running try to run now.
                reservation.attr['starttime'] = str(time.time())
                scheduled = 0
                while( not scheduled ):
                    nodelist = []
                    for job in self.fifo: # add up all the nodes being used during this job time
                        if reservation.runsDuring(job) and job.attr['status'] == "running":
                            nodelist += job.getNodelist()
                    newnodelist = self.mynodes.enoughNodes( int(reservation.attr['nodes']), reservation.attr['attribute'], nodelist )
                    if len(newnodelist):
		    	syslog(LOG_INFO,"jobid: %s nodelist: %s"%(reservation.attr['jobid'],commastring.join(newnodelist)) )
                        reservation.setNodelist(newnodelist)
			reservation.attr['status'] = "scheduled"
                        self.fifo.append(reservation)
                        scheduled = 1
                    elif self.mynodes.maxNodes() >= int(reservation.attr['nodes']): #is job small enough to ever run
                        endfound = 0 
                        counter = 0
                        while( not endfound ):
                            if counter < len(self.fifo): # find a new starttime to use
                                if reservation.getEndTime()<= self.fifo[counter].getStartTime():
                                    reservation.attr['starttime'] = self.fifo[counter].attr['starttime']
                                    endfound = 1
                            else:
                                reservation.attr['starttime'] = str( self.fifo[counter - 1].getEndTime() ) # put it at the end
				reservation.attr['status'] = "scheduled"
                                self.fifo.append(reservation)
                                endfound = 1
                                scheduled = 1
                            counter = counter +1
                    else:
                        syslog( LOG_INFO , "Not scheduling Job: %s  wants: %s nodes  total: %s nodes"
                                %(reservation.attr['jobid'],reservation.attr['nodes'], self.mynodes.maxNodes()))
                        scheduled = 1 
        
if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from sys import argv
    try:
        (opts, args) = getopt(argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nsched.py [-d] [--daemon <pidfile>]"% (msg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = Scheduler(debug=debug)
    server.ServeForever()
